@@ -2,7 +2,7 @@
 [styling=C]
 
 [keywords]
-primary=abstract assert break case catch class const continue default do else enum extends final finally for future generic goto if implements import inner instanceof interface native new outer package private protected public rest return static super switch synchronized this throw throws transient try var volatile while true false null
+primary=abstract assert break case catch class const continue default do else enum extends final finally for future generic goto if implements import inner instanceof interface native new outer package private protected public rest return static strictfp super switch synchronized this throw throws transient try var volatile while true false null
 secondary=boolean byte char double float int long short void
 # documentation keywords for javadoc
 doccomment=author deprecated exception param return see serial serialData serialField since throws todo version
@@ -13,6 +13,9 @@ typedefs=
 # default extension used when saving files
 extension=java
 
+# MIME type
+mime_type=text/x-java
+
 # the following characters are these which a "word" can contains, see documentation
 #wordchars=_abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789
 
