@@ -14,16 +14,21 @@
  *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  *      GNU General Public License for more details.
  *
- *      You should have received a copy of the GNU General Public License
- *      along with this program; if not, write to the Free Software
- *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
+ *      You should have received a copy of the GNU General Public License along
+ *      with this program; if not, write to the Free Software Foundation, Inc.,
+ *      51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  */
 
 
-#ifndef GEANY_ABOUT_DIALOG_H
-#define GEANY_ABOUT_DIALOG_H
+#ifndef GEANY_ABOUT_H
+#define GEANY_ABOUT_H 1
 
+#include <glib.h>
+
+G_BEGIN_DECLS
 
 void about_dialog_show(void);
 
-#endif
+G_END_DECLS
+
+#endif /* GEANY_ABOUT_H */
