@@ -23,81 +23,59 @@
  * externs and function prototypes are implemented in main.c. */
 
 #ifndef GEANY_H
-#define GEANY_H
+#define GEANY_H 1
+
+/* This is included here for compatibility with when GeanyApp used to be
+ * defined in this header. Some plugins (ex. GeanyLua) include individual
+ * headers instead of geanyplugin.h for some reason so they wouldn't
+ * get the GeanyApp definition if this isn't here. */
+#include "app.h"
+
+#include <glib.h>
 
-#include <gtk/gtk.h>
 
 G_BEGIN_DECLS
 
-#if defined(HAVE_CONFIG_H) && defined(GEANY_PRIVATE)
-#	include "config.h"
-#endif
-
-#include "tm_tagmanager.h"
-
-#ifndef PLAT_GTK
-#   define PLAT_GTK 1	/* needed when including ScintillaWidget.h */
-#endif
-
-/* Compatibility for sharing macros between API and core, overridden in plugindata.h */
-#define GEANY(symbol_name) symbol_name
-
 
 /* for detailed description look in the documentation, things are not
  * listed in the documentation should not be changed */
 #define GEANY_FILEDEFS_SUBDIR			"filedefs"
 #define GEANY_TEMPLATES_SUBDIR			"templates"
-#define GEANY_CODENAME					"Traeda"
+#define GEANY_CODENAME					"Rezer"
 #define GEANY_HOMEPAGE					"http://www.geany.org/"
 #define GEANY_WIKI						"http://wiki.geany.org/"
 #define GEANY_BUG_REPORT				"http://www.geany.org/Support/Bugs"
-#define GEANY_DONATE					"https://sourceforge.net/donate/index.php?group_id=153444"
+#define GEANY_DONATE					"http://www.geany.org/service/donate/"
 #define GEANY_STRING_UNTITLED			_("untitled")
 #define GEANY_DEFAULT_DIALOG_HEIGHT		350
 #define GEANY_WINDOW_DEFAULT_WIDTH		900
 #define GEANY_WINDOW_DEFAULT_HEIGHT		600
 
-
-
-/* Common forward declarations */
-typedef struct GeanyDocument GeanyDocument;
-typedef struct GeanyEditor GeanyEditor;
-typedef struct GeanyFiletype GeanyFiletype;
-
-
-/** Important application fields. */
-typedef struct GeanyApp
-{
-	gboolean			debug_mode;		/**< @c TRUE if debug messages should be printed. */
-	/** User configuration directory, usually @c ~/.config/geany.
-	 * This is a full path read by @ref tm_get_real_path().
-	 * @note Plugin configuration files should be saved as:
-	 * @code g_build_path(G_DIR_SEPARATOR_S, geany->app->configdir, "plugins", "pluginname",
-	 * 	"file.conf", NULL); @endcode */
-	gchar				*configdir;
-	gchar				*datadir;
-	gchar				*docdir;
-	const TMWorkspace	*tm_workspace;	/**< TagManager workspace/session tags. */
-	struct GeanyProject	*project;		/**< Currently active project or @c NULL if none is open. */
-}
-GeanyApp;
-
-extern GeanyApp *app;
-
-extern GObject *geany_object;
-
-
-extern gboolean	ignore_callback;
-
-
-/* prototype is here so that all files can use it. */
-void geany_debug(gchar const *format, ...) G_GNUC_PRINTF (1, 2);
-
-
 #ifndef G_GNUC_WARN_UNUSED_RESULT
 #define G_GNUC_WARN_UNUSED_RESULT
 #endif
 
+#if defined(GEANY_PRIVATE) || defined(GEANY_DISABLE_DEPRECATION_WARNINGS)
+#	define GEANY_DEPRECATED
+#	define GEANY_DEPRECATED_FOR(x)
+#else
+#	define GEANY_DEPRECATED			G_GNUC_DEPRECATED
+#	define GEANY_DEPRECATED_FOR(x)	G_GNUC_DEPRECATED_FOR(x)
+#endif
+
+/* Re-defined by plugindata.h as something else */
+#ifndef GEANY
+# define GEANY(symbol_name) symbol_name
+#endif
+
+
+#ifdef GEANY_PRIVATE
+
+/* prototype is here so that all files can use it. */
+void geany_debug(gchar const *format, ...) G_GNUC_PRINTF (1, 2);
+
+#endif /* GEANY_PRIVATE */
+
 G_END_DECLS
 
-#endif
+#endif /* GEANY_H */
