@@ -32,7 +32,10 @@
  * Emitted just after loading main keyfile settings.
  */
 
-#include "geany.h"
+#ifdef HAVE_CONFIG_H
+# include "config.h"
+#endif
+
 #include "geanyobject.h"
 
 /* extern in geany.h */
