@@ -14,17 +14,23 @@
  *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  *      GNU General Public License for more details.
  *
- *      You should have received a copy of the GNU General Public License
- *      along with this program; if not, write to the Free Software
- *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
- *      MA 02110-1301, USA.
+ *      You should have received a copy of the GNU General Public License along
+ *      with this program; if not, write to the Free Software Foundation, Inc.,
+ *      51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  */
 
 
-#ifndef GEANYOBJECT_H
-#define GEANYOBJECT_H
+#ifndef GEANY_OBJECT_H
+#define GEANY_OBJECT_H 1
+
+#include "document.h"
+#include "editor.h"
+#include "filetypes.h"
+
+#include "Scintilla.h"
+
+#include "gtkcompat.h"
 
-#include <glib-object.h>
 
 G_BEGIN_DECLS
 
@@ -74,7 +80,7 @@ struct _GeanyObject
 	/* add your public declarations here */
 };
 
-struct SCNotification;
+extern GObject *geany_object;
 
 struct _GeanyObjectClass
 {
@@ -107,4 +113,4 @@ GObject*	geany_object_new		(void);
 
 G_END_DECLS
 
-#endif /* GEANYOBJECT_H */
+#endif /* GEANY_OBJECT_H */
