@@ -51,11 +51,26 @@
  *
  */
 
-
-#include "geany.h"
+#ifdef HAVE_CONFIG_H
+# include "config.h"
+#endif
 
 #ifdef HAVE_SOCKET
 
+#include "socket.h"
+
+#include "app.h"
+#include "dialogs.h"
+#include "document.h"
+#include "encodings.h"
+#include "main.h"
+#include "support.h"
+#include "utils.h"
+#include "win32.h"
+
+#include "gtkcompat.h"
+
+
 #ifndef G_OS_WIN32
 # include <sys/time.h>
 # include <sys/types.h>
@@ -78,16 +93,6 @@
 #include <gdk/gdkx.h>
 #endif
 
-#include "main.h"
-#include "socket.h"
-#include "document.h"
-#include "support.h"
-#include "ui_utils.h"
-#include "utils.h"
-#include "dialogs.h"
-#include "encodings.h"
-#include "project.h"
-
 
 #ifdef G_OS_WIN32
 #define REMOTE_CMD_PORT		49876
@@ -98,7 +103,7 @@
 #endif
 #define BUFFER_LENGTH 4096
 
-struct socket_info_struct socket_info;
+struct SocketInfo socket_info;
 
 
 #ifdef G_OS_WIN32
@@ -205,7 +210,7 @@ static void socket_get_document_list(gint sock)
 
 	do
 	{
-		n_read = socket_fd_read(sock, buf, sizeof(buf));
+		n_read = socket_fd_read(sock, buf, BUFFER_LENGTH);
 		/* if we received ETX (end-of-text), there is nothing else to read, so cut that
 		 * byte not to output it and to be sure not to validate the loop condition */
 		if (n_read > 0 && buf[n_read - 1] == '\3')
@@ -213,14 +218,14 @@ static void socket_get_document_list(gint sock)
 		if (n_read > 0)
 			fwrite(buf, 1, n_read, stdout);
 	}
-	while (n_read >= sizeof(buf));
+	while (n_read >= BUFFER_LENGTH);
 }
 
 
 #ifndef G_OS_WIN32
 static void check_socket_permissions(void)
 {
-	struct stat socket_stat;
+	GStatBuf socket_stat;
 
 	if (g_lstat(socket_info.file_name, &socket_stat) == 0)
 	{	/* If the user id of the process is not the same as the owner of the socket
@@ -275,10 +280,13 @@ gint socket_init(gint argc, gchar **argv)
 	if (sock < 0)
 		return -1;
 #else
-	gchar *display_name = gdk_get_display();
+	gchar *display_name = NULL;
 	const gchar *hostname = g_get_host_name();
+	GdkDisplay *display = gdk_display_get_default();
 	gchar *p;
 
+	if (display != NULL)
+		display_name = g_strdup(gdk_display_get_name(display));
 	if (display_name == NULL)
 		display_name = g_strdup("NODISPLAY");
 
