@@ -68,6 +68,10 @@
 # include "support.h"
 #endif
 
+#if ! GLIB_CHECK_VERSION(2, 31, 20) && ! defined(G_SPAWN_ERROR_TOO_BIG)
+# define G_SPAWN_ERROR_TOO_BIG G_SPAWN_ERROR_2BIG
+#endif
+
 #ifdef G_OS_WIN32
 /* Each 4KB under Windows seem to come in 2 portions, so 2K + 2K is more
    balanced than 4095 + 1. May be different on the latest Windows/glib? */
@@ -701,11 +705,11 @@ static gboolean spawn_async_with_pipes(const gchar *working_directory, const gch
  *
  *  If a @a child_pid is passed, it's your responsibility to invoke @c g_spawn_close_pid().
  *
- *  @param working_directory child's current working directory, or @c NULL.
- *  @param command_line child program and arguments, or @c NULL.
- *  @param argv child's argument vector, or @c NULL.
- *  @param envp child's environment, or @c NULL.
- *  @param child_pid return location for child process ID, or @c NULL.
+ *  @param working_directory @nullable child's current working directory, or @c NULL.
+ *  @param command_line @nullable child program and arguments, or @c NULL.
+ *  @param argv @nullable child's argument vector, or @c NULL.
+ *  @param envp @nullable child's environment, or @c NULL.
+ *  @param child_pid @out @optional return location for child process ID, or @c NULL.
  *  @param error return location for error.
  *
  *  @return @c TRUE on success, @c FALSE on error.
@@ -946,7 +950,7 @@ static void spawn_watch_cb(GPid pid, gint status, gpointer data)
 }
 
 
-/**
+/** @girskip
  *  Executes a child program and setups callbacks.
  *
  *  A command line or an argument vector must be passed. If both are present, the argument
@@ -974,22 +978,22 @@ static void spawn_watch_cb(GPid pid, gint status, gpointer data)
  *
  *  The @a child_pid will be closed automatically, after @a exit_cb is invoked.
  *
- *  @param working_directory child's current working directory, or @c NULL.
- *  @param command_line child program and arguments, or @c NULL.
- *  @param argv child's argument vector, or @c NULL.
- *  @param envp child's environment, or @c NULL.
+ *  @param working_directory @nullable child's current working directory, or @c NULL.
+ *  @param command_line @nullable child program and arguments, or @c NULL.
+ *  @param argv @nullable child's argument vector, or @c NULL.
+ *  @param envp @nullable child's environment, or @c NULL.
  *  @param spawn_flags flags from SpawnFlags.
- *  @param stdin_cb callback to send data to childs's stdin, or @c NULL.
+ *  @param stdin_cb @nullable callback to send data to childs's stdin, or @c NULL.
  *  @param stdin_data data to pass to @a stdin_cb.
- *  @param stdout_cb callback to receive child's stdout, or @c NULL.
+ *  @param stdout_cb @nullable callback to receive child's stdout, or @c NULL.
  *  @param stdout_data data to pass to @a stdout_cb.
  *  @param stdout_max_length maximum data length to pass to stdout_cb, @c 0 = default.
- *  @param stderr_cb callback to receive child's stderr, or @c NULL.
+ *  @param stderr_cb @nullable callback to receive child's stderr, or @c NULL.
  *  @param stderr_data data to pass to @a stderr_cb.
  *  @param stderr_max_length maximum data length to pass to stderr_cb, @c 0 = default.
- *  @param exit_cb callback to invoke when the child exits, or @c NULL.
+ *  @param exit_cb @nullable callback to invoke when the child exits, or @c NULL.
  *  @param exit_data data to pass to @a exit_cb.
- *  @param child_pid return location for child process ID, or @c NULL.
+ *  @param child_pid @out @optional return location for child process ID, or @c NULL.
  *  @param error return location for error.
  *
  *  @return @c TRUE on success, @c FALSE on error.
@@ -1179,14 +1183,14 @@ static void spawn_get_exit_status_cb(G_GNUC_UNUSED GPid pid, gint status, gpoint
  *  All output from the child, including the nul characters, is stored in @a stdout_data and
  *  @a stderr_data (if non-NULL). Any existing data in these strings will be erased.
  *
- *  @param working_directory child's current working directory, or @c NULL.
- *  @param command_line child program and arguments, or @c NULL.
- *  @param argv child's argument vector, or @c NULL.
- *  @param envp child's environment, or @c NULL.
- *  @param stdin_data data to send to childs's stdin, or @c NULL.
- *  @param stdout_data GString location to receive the child's stdout, or NULL.
- *  @param stderr_data GString location to receive the child's stderr, or NULL.
- *  @param exit_status return location for the child exit code, or NULL.
+ *  @param working_directory @nullable child's current working directory, or @c NULL.
+ *  @param command_line @nullable child program and arguments, or @c NULL.
+ *  @param argv @nullable child's argument vector, or @c NULL.
+ *  @param envp @nullable child's environment, or @c NULL.
+ *  @param stdin_data @nullable data to send to childs's stdin, or @c NULL.
+ *  @param stdout_data @nullable GString location to receive the child's stdout, or @c NULL.
+ *  @param stderr_data @nullable GString location to receive the child's stderr, or @c NULL.
+ *  @param exit_status @out @optional return location for the child exit code, or @c NULL.
  *  @param error return location for error.
  *
  *  @return @c TRUE on success, @c FALSE on error.
@@ -1198,8 +1202,10 @@ gboolean spawn_sync(const gchar *working_directory, const gchar *command_line, g
 	gchar **envp, SpawnWriteData *stdin_data, GString *stdout_data, GString *stderr_data,
 	gint *exit_status, GError **error)
 {
-	g_string_truncate(stdout_data, 0);
-	g_string_truncate(stderr_data, 0);
+	if (stdout_data)
+		g_string_truncate(stdout_data, 0);
+	if (stderr_data)
+		g_string_truncate(stderr_data, 0);
 
 	return spawn_with_callbacks(working_directory, command_line, argv, envp, SPAWN_SYNC |
 		SPAWN_UNBUFFERED, stdin_data ? (GIOFunc) spawn_write_data : NULL, stdin_data,
