@@ -23,6 +23,12 @@
 #ifndef GEANY_TOOLS_H
 #define GEANY_TOOLS_H 1
 
+#include "document.h"
+
+#include <glib.h>
+
+G_BEGIN_DECLS
+
 void tools_create_insert_custom_command_menu_items(void);
 
 void tools_execute_custom_command(GeanyDocument *doc, const gchar *command);
@@ -31,4 +37,6 @@ void tools_word_count(void);
 
 void tools_color_chooser(const gchar *color);
 
-#endif
+G_END_DECLS
+
+#endif /* GEANY_TOOLS_H */
