@@ -23,7 +23,24 @@
  * General utility functions, non-GTK related.
  */
 
-#include "geany.h"
+#ifdef HAVE_CONFIG_H
+# include "config.h"
+#endif
+
+#include "utils.h"
+
+#include "app.h"
+#include "dialogs.h"
+#include "document.h"
+#include "prefs.h"
+#include "prefix.h"
+#include "sciwrappers.h"
+#include "spawn.h"
+#include "support.h"
+#include "templates.h"
+#include "ui_utils.h"
+#include "win32.h"
+#include "osx.h"
 
 #include <stdlib.h>
 #include <ctype.h>
@@ -41,20 +58,8 @@
 #endif
 
 #include <glib/gstdio.h>
-
 #include <gio/gio.h>
 
-#include "prefs.h"
-#include "support.h"
-#include "document.h"
-#include "filetypes.h"
-#include "dialogs.h"
-#include "win32.h"
-#include "project.h"
-#include "ui_utils.h"
-
-#include "utils.h"
-
 
 /**
  *  Tries to open the given URI in a browser.
@@ -66,35 +71,28 @@
  *
  *  @since 0.16
  **/
+GEANY_API_SYMBOL
 void utils_open_browser(const gchar *uri)
 {
 #ifdef G_OS_WIN32
 	g_return_if_fail(uri != NULL);
 	win32_open_browser(uri);
 #else
-	gboolean again = TRUE;
+	gchar *argv[2] = { (gchar *) uri, NULL };
 
 	g_return_if_fail(uri != NULL);
 
-	while (again)
+	while (!spawn_async(NULL, tool_prefs.browser_cmd, argv, NULL, NULL, NULL))
 	{
-		gchar *cmdline = g_strconcat(tool_prefs.browser_cmd, " \"", uri, "\"", NULL);
+		gchar *new_cmd = dialogs_show_input(_("Select Browser"), GTK_WINDOW(main_widgets.window),
+			_("Failed to spawn the configured browser command. "
+			  "Please correct it or enter another one."),
+			tool_prefs.browser_cmd);
 
-		if (g_spawn_command_line_async(cmdline, NULL))
-			again = FALSE;
-		else
-		{
-			gchar *new_cmd = dialogs_show_input(_("Select Browser"), GTK_WINDOW(main_widgets.window),
-				_("Failed to spawn the configured browser command. "
-				  "Please correct it or enter another one."),
-				tool_prefs.browser_cmd);
+		if (new_cmd == NULL) /* user canceled */
+			break;
 
-			if (new_cmd == NULL) /* user canceled */
-				again = FALSE;
-			else
-				SETPTR(tool_prefs.browser_cmd, new_cmd);
-		}
-		g_free(cmdline);
+		SETPTR(tool_prefs.browser_cmd, new_cmd);
 	}
 #endif
 }
@@ -207,6 +205,7 @@ gboolean utils_is_opening_brace(gchar c, gboolean include_angles)
  * @return 0 if the file was successfully written, otherwise the @c errno of the
  *         failed operation is returned.
  **/
+GEANY_API_SYMBOL
 gint utils_write_file(const gchar *filename, const gchar *text)
 {
 	g_return_val_if_fail(filename != NULL, ENOENT);
@@ -254,7 +253,7 @@ gint utils_write_file(const gchar *filename, const gchar *text)
 		{
 			geany_debug("utils_write_file(): could not write to file %s (%s)",
 				filename, g_strerror(errno));
-			return NVL(errno, EIO);
+			return FALLBACK(errno, EIO);
 		}
 	}
 	return 0;
@@ -264,8 +263,9 @@ gint utils_write_file(const gchar *filename, const gchar *text)
 /** Searches backward through @a size bytes looking for a '<'.
  * @param sel .
  * @param size .
- * @return The tag name (newly allocated) or @c NULL if no opening tag was found.
+ * @return @nullable The tag name (newly allocated) or @c NULL if no opening tag was found.
  */
+GEANY_API_SYMBOL
 gchar *utils_find_open_xml_tag(const gchar sel[], gint size)
 {
 	const gchar *cur, *begin;
@@ -288,8 +288,9 @@ gchar *utils_find_open_xml_tag(const gchar sel[], gint size)
 /** Searches backward through @a size bytes looking for a '<'.
  * @param sel .
  * @param size .
- * @return pointer to '<' of the found opening tag within @a sel, or @c NULL if no opening tag was found.
+ * @return @nullable pointer to '<' of the found opening tag within @a sel, or @c NULL if no opening tag was found.
  */
+GEANY_API_SYMBOL
 const gchar *utils_find_open_xml_tag_pos(const gchar sel[], gint size)
 {
 	/* stolen from anjuta and modified */
@@ -371,13 +372,24 @@ const gchar *utils_get_eol_name(gint eol_mode)
 {
 	switch (eol_mode)
 	{
-		case SC_EOL_CRLF: return _("Win (CRLF)"); break;
-		case SC_EOL_CR: return _("Mac (CR)"); break;
+		case SC_EOL_CRLF: return _("Windows (CRLF)"); break;
+		case SC_EOL_CR: return _("Classic Mac (CR)"); break;
 		default: return _("Unix (LF)"); break;
 	}
 }
 
 
+const gchar *utils_get_eol_short_name(gint eol_mode)
+{
+	switch (eol_mode)
+	{
+		case SC_EOL_CRLF: return _("CRLF"); break;
+		case SC_EOL_CR: return _("CR"); break;
+		default: return _("LF"); break;
+	}
+}
+
+
 const gchar *utils_get_eol_char(gint eol_mode)
 {
 	switch (eol_mode)
@@ -419,7 +431,7 @@ gboolean utils_atob(const gchar *str)
 /* NULL-safe version of g_path_is_absolute(). */
 gboolean utils_is_absolute_path(const gchar *path)
 {
-	if (G_UNLIKELY(! NZV(path)))
+	if (G_UNLIKELY(EMPTY(path)))
 		return FALSE;
 
 	return g_path_is_absolute(path);
@@ -478,14 +490,15 @@ static gchar *utf8_strdown(const gchar *str)
  *
  *  The input strings should be in UTF-8 or locale encoding.
  *
- *  @param s1 Pointer to first string or @c NULL.
- *  @param s2 Pointer to second string or @c NULL.
+ *  @param s1 @nullable Pointer to first string or @c NULL.
+ *  @param s2 @nullable Pointer to second string or @c NULL.
  *
  *  @return an integer less than, equal to, or greater than zero if @a s1 is found, respectively,
  *          to be less than, to match, or to be greater than @a s2.
  *
  *  @since 0.16
  */
+GEANY_API_SYMBOL
 gint utils_str_casecmp(const gchar *s1, const gchar *s2)
 {
 	gchar *tmp1, *tmp2;
@@ -528,6 +541,7 @@ gint utils_str_casecmp(const gchar *s1, const gchar *s2)
  *  @since 0.17
  */
 /* This following function is taken from Gedit. */
+GEANY_API_SYMBOL
 gchar *utils_str_middle_truncate(const gchar *string, guint truncate_length)
 {
 	GString *truncated;
@@ -544,7 +558,7 @@ gchar *utils_str_middle_truncate(const gchar *string, guint truncate_length)
 
 	g_return_val_if_fail(g_utf8_validate(string, length, NULL), NULL);
 
-	/* It doesnt make sense to truncate strings to less than the size of the delimiter plus 2
+	/* It doesn't make sense to truncate strings to less than the size of the delimiter plus 2
 	 * characters (one on each side) */
 	delimiter_length = g_utf8_strlen(delimiter, -1);
 	if (truncate_length < (delimiter_length + 2))
@@ -572,22 +586,19 @@ gchar *utils_str_middle_truncate(const gchar *string, guint truncate_length)
  *  @c NULL-safe string comparison. Returns @c TRUE if both @a a and @a b are @c NULL
  *  or if @a a and @a b refer to valid strings which are equal.
  *
- *  @param a Pointer to first string or @c NULL.
- *  @param b Pointer to second string or @c NULL.
+ *  @param a @nullable Pointer to first string or @c NULL.
+ *  @param b @nullable Pointer to second string or @c NULL.
  *
  *  @return @c TRUE if @a a equals @a b, else @c FALSE.
  **/
+GEANY_API_SYMBOL
 gboolean utils_str_equal(const gchar *a, const gchar *b)
 {
 	/* (taken from libexo from os-cillation) */
 	if (a == NULL && b == NULL) return TRUE;
 	else if (a == NULL || b == NULL) return FALSE;
 
-	while (*a == *b++)
-		if (*a++ == '\0')
-			return TRUE;
-
-	return FALSE;
+	return strcmp(a, b) == 0;
 }
 
 
@@ -598,6 +609,7 @@ gboolean utils_str_equal(const gchar *a, const gchar *b)
  *
  *  @return A newly-allocated string, should be freed when no longer needed.
  **/
+GEANY_API_SYMBOL
 gchar *utils_remove_ext_from_filename(const gchar *filename)
 {
 	gchar *last_dot;
@@ -636,19 +648,6 @@ gchar utils_brace_opposite(gchar ch)
 }
 
 
-gchar *utils_get_hostname(void)
-{
-#ifdef G_OS_WIN32
-	return win32_get_hostname();
-#elif defined(HAVE_GETHOSTNAME)
-	gchar hostname[100];
-	if (gethostname(hostname, sizeof(hostname)) == 0)
-		return g_strdup(hostname);
-#endif
-	return g_strdup("localhost");
-}
-
-
 /* Checks whether the given file can be written. locale_filename is expected in locale encoding.
  * Returns 0 if it can be written, otherwise it returns errno */
 gint utils_is_file_writable(const gchar *locale_filename)
@@ -723,12 +722,13 @@ gint utils_strpos(const gchar *haystack, const gchar *needle)
  *
  *  @param format The format string to pass to strftime(3). See the strftime(3)
  *                documentation for details, in UTF-8 encoding.
- *  @param time_to_use The date/time to use, in time_t format or NULL to use the current time.
+ *  @param time_to_use @nullable The date/time to use, in time_t format or @c NULL to use the current time.
  *
  *  @return A newly-allocated string, should be freed when no longer needed.
  *
  *  @since 0.16
  */
+GEANY_API_SYMBOL
 gchar *utils_get_date_time(const gchar *format, time_t *time_to_use)
 {
 	const struct tm *tm;
@@ -797,6 +797,7 @@ gchar *utils_get_initials(const gchar *name)
  *  @return The value associated with @a key as an integer, or the given default value if the value
  *          could not be retrieved.
  **/
+GEANY_API_SYMBOL
 gint utils_get_setting_integer(GKeyFile *config, const gchar *section, const gchar *key,
 							   const gint default_value)
 {
@@ -827,6 +828,7 @@ gint utils_get_setting_integer(GKeyFile *config, const gchar *section, const gch
  *  @return The value associated with @a key as a boolean, or the given default value if the value
  *          could not be retrieved.
  **/
+GEANY_API_SYMBOL
 gboolean utils_get_setting_boolean(GKeyFile *config, const gchar *section, const gchar *key,
 								   const gboolean default_value)
 {
@@ -857,6 +859,7 @@ gboolean utils_get_setting_boolean(GKeyFile *config, const gchar *section, const
  *  @return A newly allocated string, either the value for @a key or a copy of the given
  *          default value if it could not be retrieved.
  **/
+GEANY_API_SYMBOL
 gchar *utils_get_setting_string(GKeyFile *config, const gchar *section, const gchar *key,
 								const gchar *default_value)
 {
@@ -875,28 +878,12 @@ gchar *utils_get_setting_string(GKeyFile *config, const gchar *section, const gc
 
 gchar *utils_get_hex_from_color(GdkColor *color)
 {
-	gchar *buffer = g_malloc0(9);
-
 	g_return_val_if_fail(color != NULL, NULL);
 
-	g_snprintf(buffer, 8, "#%02X%02X%02X",
-	      (guint) (utils_scale_round(color->red / 256, 255)),
-	      (guint) (utils_scale_round(color->green / 256, 255)),
-	      (guint) (utils_scale_round(color->blue / 256, 255)));
-
-	return buffer;
-}
-
-
-guint utils_invert_color(guint color)
-{
-	guint r, g, b;
-
-	r = 0xffffff - color;
-	g = 0xffffff - (color >> 8);
-	b = 0xffffff - (color >> 16);
-
-	return (r | (g << 8) | (b << 16));
+	return g_strdup_printf("#%02X%02X%02X",
+		(guint) (utils_scale_round(color->red / 256, 255)),
+		(guint) (utils_scale_round(color->green / 256, 255)),
+		(guint) (utils_scale_round(color->blue / 256, 255)));
 }
 
 
@@ -980,50 +967,44 @@ gchar *utils_make_human_readable_str(guint64 size, gulong block_size,
 }
 
 
- static guint utils_get_value_of_hex(const gchar ch)
+/* converts a color representation using gdk_color_parse(), with additional
+ * support of the "0x" prefix as a synonym for "#" */
+gboolean utils_parse_color(const gchar *spec, GdkColor *color)
 {
-	if (ch >= '0' && ch <= '9')
-		return ch - '0';
-	else if (ch >= 'A' && ch <= 'F')
-		return ch - 'A' + 10;
-	else if (ch >= 'a' && ch <= 'f')
-		return ch - 'a' + 10;
-	else
-		return 0;
+	gchar buf[64] = {0};
+
+	g_return_val_if_fail(spec != NULL, -1);
+
+	if (spec[0] == '0' && (spec[1] == 'x' || spec[1] == 'X'))
+	{
+		/* convert to # format for GDK to understand it */
+		buf[0] = '#';
+		strncpy(buf + 1, spec + 2, sizeof(buf) - 2);
+		spec = buf;
+	}
+
+	return gdk_color_parse(spec, color);
 }
 
 
-/* utils_strtod() converts a string containing a hex colour ("0x00ff00") into an integer.
- * Basically, it is the same as strtod() would do, but it does not understand hex colour values,
- * before ANSI-C99. With with_route set, it takes strings of the format "#00ff00".
- * Returns -1 on failure. */
-gint utils_strtod(const gchar *source, gchar **end, gboolean with_route)
+/* converts a GdkColor to the packed 24 bits BGR format, as understood by Scintilla
+ * returns a 24 bits BGR color, or -1 on failure */
+gint utils_color_to_bgr(const GdkColor *c)
 {
-	guint red, green, blue, offset = 0;
+	g_return_val_if_fail(c != NULL, -1);
+	return (c->red / 256) | ((c->green / 256) << 8) | ((c->blue / 256) << 16);
+}
 
-	g_return_val_if_fail(source != NULL, -1);
 
-	if (with_route && (strlen(source) != 7 || source[0] != '#'))
+/* parses @p spec using utils_parse_color() and convert it to 24 bits BGR using
+ * utils_color_to_bgr() */
+gint utils_parse_color_to_bgr(const gchar *spec)
+{
+	GdkColor color;
+	if (utils_parse_color(spec, &color))
+		return utils_color_to_bgr(&color);
+	else
 		return -1;
-	else if (! with_route && (strlen(source) != 8 || source[0] != '0' ||
-		(source[1] != 'x' && source[1] != 'X')))
-	{
-		return -1;
-	}
-
-	/* offset is set to 1 when the string starts with 0x, otherwise it starts with #
-	 * and we don't need to increase the index */
-	if (! with_route)
-		offset = 1;
-
-	red = utils_get_value_of_hex(
-					source[1 + offset]) * 16 + utils_get_value_of_hex(source[2 + offset]);
-	green = utils_get_value_of_hex(
-					source[3 + offset]) * 16 + utils_get_value_of_hex(source[4 + offset]);
-	blue = utils_get_value_of_hex(
-					source[5 + offset]) * 16 + utils_get_value_of_hex(source[6 + offset]);
-
-	return (red | (green << 8) | (blue << 16));
 }
 
 
@@ -1272,6 +1253,7 @@ gboolean utils_wrap_string(gchar *string, gint wrapstart)
  *  @return The converted string in locale encoding, or a copy of the input string if conversion
  *    failed. Should be freed with g_free(). If @a utf8_text is @c NULL, @c NULL is returned.
  **/
+GEANY_API_SYMBOL
 gchar *utils_get_locale_from_utf8(const gchar *utf8_text)
 {
 #ifdef G_OS_WIN32
@@ -1300,6 +1282,7 @@ gchar *utils_get_locale_from_utf8(const gchar *utf8_text)
  *  @return The converted string in UTF-8 encoding, or a copy of the input string if conversion
  *    failed. Should be freed with g_free(). If @a locale_text is @c NULL, @c NULL is returned.
  **/
+GEANY_API_SYMBOL
 gchar *utils_get_utf8_from_locale(const gchar *locale_text)
 {
 #ifdef G_OS_WIN32
@@ -1389,6 +1372,7 @@ gchar **utils_strv_new(const gchar *first, ...)
  *  @return 0 if the directory was successfully created, otherwise the @c errno of the
  *          failed operation is returned.
  **/
+GEANY_API_SYMBOL
 gint utils_mkdir(const gchar *path, gboolean create_parent_dirs)
 {
 	gint mode = 0700;
@@ -1421,10 +1405,11 @@ gint utils_mkdir(const gchar *path, gboolean create_parent_dirs)
  * @param sort Whether to sort alphabetically (UTF-8 safe).
  * @param error The location for storing a possible error, or @c NULL.
  *
- * @return A newly allocated list or @c NULL if no files were found. The list and its data should be
- * freed when no longer needed.
+ * @return @elementtype{filename} @transfer{full} @nullable A newly allocated list or @c NULL if
+ * no files were found. The list and its data should be freed when no longer needed.
  * @see utils_get_file_list().
  **/
+GEANY_API_SYMBOL
 GSList *utils_get_file_list_full(const gchar *path, gboolean full_path, gboolean sort, GError **error)
 {
 	GSList *list = NULL;
@@ -1465,10 +1450,11 @@ GSList *utils_get_file_list_full(const gchar *path, gboolean full_path, gboolean
  *               unless @c NULL.
  * @param error The location for storing a possible error, or @c NULL.
  *
- * @return A newly allocated list or @c NULL if no files were found. The list and its data should be
- *         freed when no longer needed.
+ * @return @elementtype{filename} @transfer{full} @nullable A newly allocated list or @c NULL
+ * if no files were found. The list and its data should be freed when no longer needed.
  * @see utils_get_file_list_full().
  **/
+GEANY_API_SYMBOL
 GSList *utils_get_file_list(const gchar *path, guint *length, GError **error)
 {
 	GSList *list = utils_get_file_list_full(path, FALSE, TRUE, error);
@@ -1484,7 +1470,7 @@ gboolean utils_str_has_upper(const gchar *str)
 {
 	gunichar c;
 
-	if (! NZV(str) || ! g_utf8_validate(str, -1, NULL))
+	if (EMPTY(str) || ! g_utf8_validate(str, -1, NULL))
 		return FALSE;
 
 	while (*str != '\0')
@@ -1514,7 +1500,7 @@ gint utils_string_find(GString *haystack, gint start, gint end, const gchar *nee
 	if (start >= (gint)haystack->len)
 		return -1;
 
-	g_return_val_if_fail(NZV(needle), -1);
+	g_return_val_if_fail(!EMPTY(needle), -1);
 
 	if (end < 0)
 		end = haystack->len;
@@ -1556,6 +1542,7 @@ gint utils_string_replace(GString *str, gint pos, gint len, const gchar *replace
  *
  * @return Number of replacements made.
  **/
+GEANY_API_SYMBOL
 guint utils_string_replace_all(GString *haystack, const gchar *needle, const gchar *replace)
 {
 	guint count = 0;
@@ -1589,6 +1576,7 @@ guint utils_string_replace_all(GString *haystack, const gchar *needle, const gch
  *
  *  @since 0.16
  */
+GEANY_API_SYMBOL
 guint utils_string_replace_first(GString *haystack, const gchar *needle, const gchar *replace)
 {
 	gint pos = utils_string_find(haystack, 0, -1, needle);
@@ -1643,12 +1631,12 @@ guint utils_string_regex_replace_all(GString *haystack, GRegex *regex,
 /* Get project or default startup directory (if set), or NULL. */
 const gchar *utils_get_default_dir_utf8(void)
 {
-	if (app->project && NZV(app->project->base_path))
+	if (app->project && !EMPTY(app->project->base_path))
 	{
 		return app->project->base_path;
 	}
 
-	if (NZV(prefs.default_open_path))
+	if (!EMPTY(prefs.default_open_path))
 	{
 		return prefs.default_open_path;
 	}
@@ -1656,107 +1644,61 @@ const gchar *utils_get_default_dir_utf8(void)
 }
 
 
-static gboolean check_error(GError **error)
-{
-	if (error != NULL && *error != NULL)
-	{
-		/* imitate the GLib warning */
-		g_warning(
-			"GError set over the top of a previous GError or uninitialized memory.\n"
-			"This indicates a bug in someone's code. You must ensure an error is NULL "
-			"before it's set.");
-		/* after returning the code may segfault, but we don't care because we should
-		 * make sure *error is NULL */
-		return FALSE;
-	}
-	return TRUE;
-}
-
-
 /**
- *  Wraps g_spawn_sync() and internally calls this function on Unix-like
- *  systems. On Win32 platforms, it uses the Windows API.
+ *  Wraps @c spawn_sync(), which see.
  *
- *  @param dir The child's current working directory, or @a NULL to inherit parent's.
+ *  @param dir @nullable The child's current working directory, or @c NULL to inherit parent's.
  *  @param argv The child's argument vector.
- *  @param env The child's environment, or @a NULL to inherit parent's.
- *  @param flags Flags from GSpawnFlags.
- *  @param child_setup A function to run in the child just before exec().
- *  @param user_data The user data for child_setup.
- *  @param std_out The return location for child output.
- *  @param std_err The return location for child error messages.
- *  @param exit_status The child exit status, as returned by waitpid().
- *  @param error The return location for error or @a NULL.
+ *  @param env @nullable The child's environment, or @c NULL to inherit parent's.
+ *  @param flags Ignored.
+ *  @param child_setup @girskip Ignored.
+ *  @param user_data @girskip Ignored.
+ *  @param std_out @out @optional The return location for child output, or @c NULL.
+ *  @param std_err @out @optional The return location for child error messages, or @c NULL.
+ *  @param exit_status @out @optional The child exit status, as returned by waitpid(), or @c NULL.
+ *  @param error The return location for error or @c NULL.
  *
  *  @return @c TRUE on success, @c FALSE if an error was set.
  **/
+GEANY_API_SYMBOL
 gboolean utils_spawn_sync(const gchar *dir, gchar **argv, gchar **env, GSpawnFlags flags,
 						  GSpawnChildSetupFunc child_setup, gpointer user_data, gchar **std_out,
 						  gchar **std_err, gint *exit_status, GError **error)
 {
-	gboolean result;
-
-	if (! check_error(error))
-		return FALSE;
-
-	if (argv == NULL)
-	{
-		*error = g_error_new(G_SPAWN_ERROR, G_SPAWN_ERROR_FAILED, "argv must not be NULL");
-		return FALSE;
-	}
+	GString *output = std_out ? g_string_new(NULL) : NULL;
+	GString *errors = std_err ? g_string_new(NULL) : NULL;
+	gboolean result = spawn_sync(dir, NULL, argv, env, NULL, output, errors, exit_status, error);
 
 	if (std_out)
-		*std_out = NULL;
+		*std_out = g_string_free(output, !result);
 
 	if (std_err)
-		*std_err = NULL;
-
-#ifdef G_OS_WIN32
-	result = win32_spawn(dir, argv, env, flags, std_out, std_err, exit_status, error);
-#else
-	result = g_spawn_sync(dir, argv, env, flags, NULL, NULL, std_out, std_err, exit_status, error);
-#endif
+		*std_err = g_string_free(errors, !result);
 
 	return result;
 }
 
 
 /**
- *  Wraps g_spawn_async() and internally calls this function on Unix-like
- *  systems. On Win32 platforms, it uses the Windows API.
+ *  Wraps @c spawn_async(), which see.
  *
- *  @param dir The child's current working directory, or @a NULL to inherit parent's.
+ *  @param dir @nullable The child's current working directory, or @c NULL to inherit parent's.
  *  @param argv The child's argument vector.
- *  @param env The child's environment, or @a NULL to inherit parent's.
- *  @param flags Flags from GSpawnFlags.
- *  @param child_setup A function to run in the child just before exec().
- *  @param user_data The user data for child_setup.
- *  @param child_pid The return location for child process ID, or NULL.
- *  @param error The return location for error or @a NULL.
+ *  @param env @nullable The child's environment, or @c NULL to inherit parent's.
+ *  @param flags Ignored.
+ *  @param child_setup @girskip Ignored.
+ *  @param user_data Ignored.
+ *  @param child_pid @nullable The return location for child process ID, or @c NULL.
+ *  @param error The return location for error or @c NULL.
  *
  *  @return @c TRUE on success, @c FALSE if an error was set.
  **/
+GEANY_API_SYMBOL
 gboolean utils_spawn_async(const gchar *dir, gchar **argv, gchar **env, GSpawnFlags flags,
 						   GSpawnChildSetupFunc child_setup, gpointer user_data, GPid *child_pid,
 						   GError **error)
 {
-	gboolean result;
-
-	if (! check_error(error))
-		return FALSE;
-
-	if (argv == NULL)
-	{
-		*error = g_error_new(G_SPAWN_ERROR, G_SPAWN_ERROR_FAILED, "argv must not be NULL");
-		return FALSE;
-	}
-
-#ifdef G_OS_WIN32
-	result = win32_spawn(dir, argv, env, flags, NULL, NULL, NULL, error);
-#else
-	result = g_spawn_async(dir, argv, env, flags, NULL, NULL, child_pid, error);
-#endif
-	return result;
+	return spawn_async(dir, NULL, argv, env, child_pid, error);
 }
 
 
@@ -1837,14 +1779,14 @@ gboolean utils_is_remote_path(const gchar *path)
  * @see tm_get_real_path() - also resolves links. */
 void utils_tidy_path(gchar *filename)
 {
-	GString *str = g_string_new(filename);
-	const gchar *c, *needle;
-	gchar *tmp;
-	gssize pos;
+	GString *str;
+	const gchar *needle;
 	gboolean preserve_double_backslash = FALSE;
 
 	g_return_if_fail(g_path_is_absolute(filename));
 
+	str = g_string_new(filename);
+
 	if (str->len >= 2 && strncmp(str->str, "\\\\", 2) == 0)
 		preserve_double_backslash = TRUE;
 
@@ -1863,11 +1805,13 @@ void utils_tidy_path(gchar *filename)
 	needle = G_DIR_SEPARATOR_S ".." G_DIR_SEPARATOR_S;
 	while (1)
 	{
-		c = strstr(str->str, needle);
+		const gchar *c = strstr(str->str, needle);
 		if (c == NULL)
 			break;
 		else
 		{
+			gssize pos, sub_len;
+
 			pos = c - str->str;
 			if (pos <= 3)
 				break;	/* bad path */
@@ -1876,17 +1820,20 @@ void utils_tidy_path(gchar *filename)
 			g_string_erase(str, pos, strlen(needle));
 			g_string_insert_c(str, pos, G_DIR_SEPARATOR);
 
-			tmp = g_strndup(str->str, pos);	/* path up to "/../" */
-			c = g_strrstr(tmp, G_DIR_SEPARATOR_S);
-			g_return_if_fail(c);
+			/* search for last "/" before found "/../" */
+			c = g_strrstr_len(str->str, pos, G_DIR_SEPARATOR_S);
+			sub_len = pos - (c - str->str);
+			if (! c)
+				break;	/* bad path */
 
-			pos = c - tmp;	/* position of previous "/" */
-			g_string_erase(str, pos, strlen(c));
-			g_free(tmp);
+			pos = c - str->str;	/* position of previous "/" */
+			g_string_erase(str, pos, sub_len);
 		}
 	}
-	g_return_if_fail(strlen(str->str) <= strlen(filename));
-	strcpy(filename, str->str);
+	if (str->len <= strlen(filename))
+		memcpy(filename, str->str, str->len + 1);
+	else
+		g_warn_if_reached();
 	g_string_free(str, TRUE);
 }
 
@@ -1902,13 +1849,14 @@ void utils_tidy_path(gchar *filename)
  *
  *  @see @c g_strdelimit.
  **/
+GEANY_API_SYMBOL
 gchar *utils_str_remove_chars(gchar *string, const gchar *chars)
 {
 	const gchar *r;
 	gchar *w = string;
 
 	g_return_val_if_fail(string, NULL);
-	if (G_UNLIKELY(! NZV(chars)))
+	if (G_UNLIKELY(EMPTY(chars)))
 		return string;
 
 	foreach_str(r, string)
@@ -1964,7 +1912,7 @@ gchar *utils_get_help_url(const gchar *suffix)
 
 #ifdef G_OS_WIN32
 	skip = 8;
-	uri = g_strconcat("file:///", app->docdir, "/Manual.html", NULL);
+	uri = g_strconcat("file:///", app->docdir, "/index.html", NULL);
 	g_strdelimit(uri, "\\", '/'); /* replace '\\' by '/' */
 #else
 	skip = 7;
@@ -2011,8 +1959,9 @@ static gboolean str_in_array(const gchar **haystack, const gchar *needle)
  * @param first_varname Name of the first variable to copy into the new array.
  * @param ... Key-value pairs of variable names and values, @c NULL-terminated.
  *
- * @return The new environment array.
+ * @return @transfer{full} The new environment array. Use @c g_strfreev() to free it.
  **/
+GEANY_API_SYMBOL
 gchar **utils_copy_environment(const gchar **exclude_vars, const gchar *first_varname, ...)
 {
 	gchar **result;
@@ -2022,9 +1971,6 @@ gchar **utils_copy_environment(const gchar **exclude_vars, const gchar *first_va
 	const gchar *key, *value;
 	guint n, o;
 
-	/* get all the environ variables */
-	env = g_listenv();
-
 	/* count the additional variables */
 	va_start(args, first_varname);
 	for (o = 1; va_arg(args, gchar*) != NULL; o++);
@@ -2034,6 +1980,9 @@ gchar **utils_copy_environment(const gchar **exclude_vars, const gchar *first_va
 
 	o /= 2;
 
+	/* get all the environ variables */
+	env = g_listenv();
+
 	/* create an array large enough to hold the new environment */
 	n = g_strv_length(env);
 	/* 'n + o + 1' could leak a little bit when exclude_vars is set */
@@ -2100,3 +2049,142 @@ gchar **utils_strv_join(gchar **first, gchar **second)
 	g_free(second);
 	return strv;
 }
+
+
+/* Try to parse a date using g_date_set_parse(). It doesn't take any format hint,
+ * obviously g_date_set_parse() uses some magic.
+ * The returned GDate object must be freed. */
+GDate *utils_parse_date(const gchar *input)
+{
+	GDate *date = g_date_new();
+
+	g_date_set_parse(date, input);
+
+	if (g_date_valid(date))
+		return date;
+
+	g_date_free(date);
+	return NULL;
+}
+
+
+gchar *utils_parse_and_format_build_date(const gchar *input)
+{
+	gchar date_buf[255];
+	GDate *date = utils_parse_date(input);
+
+	if (date != NULL)
+	{
+		g_date_strftime(date_buf, sizeof(date_buf), GEANY_TEMPLATES_FORMAT_DATE, date);
+		g_date_free(date);
+		return g_strdup(date_buf);
+	}
+
+	return g_strdup(input);
+}
+
+
+gchar *utils_get_user_config_dir(void)
+{
+#ifdef G_OS_WIN32
+	return win32_get_user_config_dir();
+#else
+	return g_build_filename(g_get_user_config_dir(), "geany", NULL);
+#endif
+}
+
+
+static gboolean is_osx_bundle(void)
+{
+#ifdef MAC_INTEGRATION
+	gchar *bundle_id = gtkosx_application_get_bundle_id();
+	if (bundle_id)
+	{
+		g_free(bundle_id);
+		return TRUE;
+	}
+#endif
+	return FALSE;
+}
+
+
+const gchar *utils_resource_dir(GeanyResourceDirType type)
+{
+	static const gchar *resdirs[RESOURCE_DIR_COUNT] = {NULL};
+
+	if (!resdirs[RESOURCE_DIR_DATA])
+	{
+#ifdef G_OS_WIN32
+		gchar *prefix = win32_get_installation_dir();
+
+		resdirs[RESOURCE_DIR_DATA] = g_build_filename(prefix, "data", NULL);
+		resdirs[RESOURCE_DIR_ICON] = g_build_filename(prefix, "share", "icons", NULL);
+		resdirs[RESOURCE_DIR_DOC] = g_build_filename(prefix, "share", "doc", "geany", "html", NULL);
+		resdirs[RESOURCE_DIR_LOCALE] = g_build_filename(prefix, "share", "locale", NULL);
+		resdirs[RESOURCE_DIR_PLUGIN] = g_build_filename(prefix, "lib", "geany", NULL);
+		g_free(prefix);
+#else
+		if (is_osx_bundle())
+		{
+# ifdef MAC_INTEGRATION
+			gchar *prefix = gtkosx_application_get_resource_path();
+			
+			resdirs[RESOURCE_DIR_DATA] = g_build_filename(prefix, "share", "geany", NULL);
+			resdirs[RESOURCE_DIR_ICON] = g_build_filename(prefix, "share", "icons", NULL);
+			resdirs[RESOURCE_DIR_DOC] = g_build_filename(prefix, "share", "doc", "geany", "html", NULL);
+			resdirs[RESOURCE_DIR_LOCALE] = g_build_filename(prefix, "share", "locale", NULL);
+			resdirs[RESOURCE_DIR_PLUGIN] = g_build_filename(prefix, "lib", "geany", NULL);
+			g_free(prefix);
+# endif
+		}
+		else
+		{
+			resdirs[RESOURCE_DIR_DATA] = g_build_filename(GEANY_DATADIR, "geany", NULL);
+			resdirs[RESOURCE_DIR_ICON] = g_build_filename(GEANY_DATADIR, "icons", NULL);
+			resdirs[RESOURCE_DIR_DOC] = g_build_filename(GEANY_DOCDIR, "html", NULL);
+			resdirs[RESOURCE_DIR_LOCALE] = g_build_filename(GEANY_LOCALEDIR, NULL);
+			resdirs[RESOURCE_DIR_PLUGIN] = g_build_filename(GEANY_LIBDIR, "geany", NULL);
+		}
+#endif
+	}
+
+	return resdirs[type];
+}
+
+
+void utils_start_new_geany_instance(const gchar *doc_path)
+{
+	const gchar *command = is_osx_bundle() ? "open" : "geany";
+	gchar *exec_path = g_find_program_in_path(command);
+
+	if (exec_path)
+	{
+		GError *err = NULL;
+		const gchar *argv[6]; // max args + 1
+		gint argc = 0;
+
+		argv[argc++] = exec_path;
+		if (is_osx_bundle())
+		{
+			argv[argc++] = "-n";
+			argv[argc++] = "-a";
+			argv[argc++] = "Geany";
+			argv[argc++] = doc_path;
+		}
+		else
+		{
+			argv[argc++] = "-i";
+			argv[argc++] = doc_path;
+		}
+		argv[argc] = NULL;
+
+		if (!utils_spawn_async(NULL, (gchar**) argv, NULL, 0, NULL, NULL, NULL, &err))
+		{
+			g_printerr("Unable to open new window: %s", err->message);
+			g_error_free(err);
+		}
+		g_free(exec_path);
+	}
+	else
+		g_printerr("Unable to find 'geany'");
+}
