@@ -76,13 +76,14 @@ typedef enum eKeywordId
 	KEYWORD_LOCAL, KEYWORD_LONG,
 	KEYWORD_M_BAD_STATE, KEYWORD_M_BAD_TRANS, KEYWORD_M_STATE, KEYWORD_M_TRANS,
 	KEYWORD_MODULE, KEYWORD_MUTABLE,
-	KEYWORD_NAMESPACE, KEYWORD_NEW, KEYWORD_NEWCOV, KEYWORD_NATIVE,
+	KEYWORD_NAMESPACE, KEYWORD_NEW, KEYWORD_NEWCOV, KEYWORD_NATIVE, KEYWORD_NOEXCEPT,
 	KEYWORD_OPERATOR, KEYWORD_OUT, KEYWORD_OUTPUT, KEYWORD_OVERLOAD, KEYWORD_OVERRIDE,
 	KEYWORD_PACKED, KEYWORD_PORT, KEYWORD_PACKAGE, KEYWORD_PRIVATE,
 	KEYWORD_PROGRAM, KEYWORD_PROTECTED, KEYWORD_PUBLIC,
 	KEYWORD_REF, KEYWORD_REGISTER, KEYWORD_RETURN,
 	KEYWORD_SHADOW, KEYWORD_STATE,
-	KEYWORD_SET, KEYWORD_SHORT, KEYWORD_SIGNAL, KEYWORD_SIGNED, KEYWORD_SIZE_T, KEYWORD_STATIC, KEYWORD_STRING,
+	KEYWORD_SET, KEYWORD_SHORT, KEYWORD_SIGNAL, KEYWORD_SIGNED, KEYWORD_SIZE_T, KEYWORD_STATIC,
+	KEYWORD_STATIC_ASSERT, KEYWORD_STRING,
 	KEYWORD_STRUCT, KEYWORD_SWITCH, KEYWORD_SYNCHRONIZED,
 	KEYWORD_TASK, KEYWORD_TEMPLATE, KEYWORD_THIS, KEYWORD_THROW,
 	KEYWORD_THROWS, KEYWORD_TRANSIENT, KEYWORD_TRANS, KEYWORD_TRANSITION,
@@ -203,6 +204,7 @@ typedef struct sStatementInfo
 	boolean			haveQualifyingName;	/* do we have a name we are considering? */
 	boolean			gotParenName;		/* was a name inside parentheses parsed yet? */
 	boolean			gotArgs;			/* was a list of parameters parsed yet? */
+	unsigned int	nSemicolons;			/* how many semicolons did we see in that statement */
 	impType			implementation;		/* abstract or concrete implementation? */
 	unsigned int	tokenIndex;			/* currently active token */
 	tokenInfo*		token [((int) NumTokens)];
@@ -326,7 +328,7 @@ typedef enum
 {
 	JK_UNDEFINED = -1,
 	JK_CLASS, JK_FIELD, JK_INTERFACE, JK_METHOD,
-	JK_PACKAGE
+	JK_PACKAGE, JK_ENUMERATOR, JK_ENUMERATION
 } javaKind;
 
 static kindOption JavaKinds [] = {
@@ -335,6 +337,8 @@ static kindOption JavaKinds [] = {
 	{ TRUE,  'i', "interface", "interfaces"},
 	{ TRUE,  'm', "method", "methods"},
 	{ TRUE,  'p', "package", "packages"},
+	{ TRUE,  'e', "enumerator", "enumerators (values inside an enumeration)"},
+	{ TRUE,  'g', "enum",       "enumeration names"},
 };
 
 typedef enum
@@ -421,7 +425,7 @@ static const keywordDesc KeywordTable [] = {
 	{ "extends",        KEYWORD_EXTENDS,        { 0, 0, 0, 1, 1, 0, 0 } },
 	{ "extern",         KEYWORD_EXTERN,         { 1, 1, 1, 0, 1, 1, 0 } },
 	{ "extern",         KEYWORD_NAMESPACE,      { 0, 0, 0, 0, 0, 0, 1 } },	/* parse block */
-	{ "final",          KEYWORD_FINAL,          { 0, 1, 0, 1, 0, 0, 1 } },
+	{ "final",          KEYWORD_FINAL,          { 0, 0, 0, 1, 0, 0, 1 } },
 	{ "finally",        KEYWORD_FINALLY,        { 0, 0, 0, 0, 0, 1, 1 } },
 	{ "float",          KEYWORD_FLOAT,          { 1, 1, 1, 1, 0, 1, 1 } },
 	{ "for",            KEYWORD_FOR,            { 1, 1, 1, 1, 0, 1, 1 } },
@@ -453,6 +457,7 @@ static const keywordDesc KeywordTable [] = {
 	{ "native",         KEYWORD_NATIVE,         { 0, 0, 0, 1, 0, 0, 0 } },
 	{ "new",            KEYWORD_NEW,            { 0, 1, 1, 1, 0, 1, 1 } },
 	{ "newcov",         KEYWORD_NEWCOV,         { 0, 0, 0, 0, 1, 0, 0 } },
+	{ "noexcept",       KEYWORD_NOEXCEPT,       { 0, 1, 0, 0, 0, 0, 0 } },
 	{ "operator",       KEYWORD_OPERATOR,       { 0, 1, 1, 0, 0, 0, 0 } },
 	{ "out",            KEYWORD_OUT,            { 0, 0, 0, 0, 0, 1, 1 } },
 	{ "output",         KEYWORD_OUTPUT,         { 0, 0, 0, 0, 1, 0, 0 } },
@@ -473,9 +478,10 @@ static const keywordDesc KeywordTable [] = {
 	{ "short",          KEYWORD_SHORT,          { 1, 1, 1, 1, 0, 1, 1 } },
 	{ "signal",         KEYWORD_SIGNAL,         { 0, 0, 0, 0, 0, 1, 0 } },
 	{ "signed",         KEYWORD_SIGNED,         { 1, 1, 0, 0, 0, 0, 0 } },
-	{ "size_t",         KEYWORD_SIZE_T,         { 1, 1, 0, 0, 0, 1, 1 } },
+	{ "size_t",         KEYWORD_SIZE_T,         { 0, 0, 0, 0, 0, 1, 0 } },
 	{ "state",          KEYWORD_STATE,          { 0, 0, 0, 0, 1, 0, 0 } },
 	{ "static",         KEYWORD_STATIC,         { 1, 1, 1, 1, 1, 1, 1 } },
+	{ "static_assert",  KEYWORD_STATIC_ASSERT,  { 0, 1, 0, 0, 0, 0, 0 } },
 	{ "string",         KEYWORD_STRING,         { 0, 0, 1, 0, 1, 1, 0 } },
 	{ "struct",         KEYWORD_STRUCT,         { 1, 1, 1, 0, 0, 1, 1 } },
 	{ "switch",         KEYWORD_SWITCH,         { 1, 1, 1, 1, 0, 1, 1 } },
@@ -501,7 +507,7 @@ static const keywordDesc KeywordTable [] = {
 	{ "virtual",        KEYWORD_VIRTUAL,        { 0, 1, 1, 0, 1, 1, 0 } },
 	{ "void",           KEYWORD_VOID,           { 1, 1, 1, 1, 1, 1, 1 } },
 	{ "volatile",       KEYWORD_VOLATILE,       { 1, 1, 1, 1, 0, 0, 1 } },
-	{ "wchar_t",        KEYWORD_WCHAR_T,        { 1, 1, 1, 0, 0, 0, 1 } },
+	{ "wchar_t",        KEYWORD_WCHAR_T,        { 0, 1, 1, 0, 0, 0, 0 } },
 	{ "weak",           KEYWORD_WEAK,           { 0, 0, 0, 0, 0, 1, 0 } },
 	{ "while",          KEYWORD_WHILE,          { 1, 1, 1, 1, 0, 1, 1 } }
 };
@@ -512,7 +518,8 @@ static const keywordDesc KeywordTable [] = {
 */
 static void createTags (const unsigned int nestLevel, statementInfo *const parent);
 static void copyToken (tokenInfo *const dest, const tokenInfo *const src);
-static const char *getVarType (const statementInfo *const st);
+static const char *getVarType (const statementInfo *const st,
+							   const tokenInfo *const token);
 
 /*
 *   FUNCTION DEFINITIONS
@@ -849,7 +856,7 @@ static boolean isContextualStatement (const statementInfo *const st)
 	{
 		if (isLanguage (Lang_vala))
 		{
-			/* All can be a contextual statment as properties can be of any type */
+			/* All can be a contextual statement as properties can be of any type */
 			result = TRUE;
 		}
 		else
@@ -938,6 +945,7 @@ static void reinitStatement (statementInfo *const st, const boolean partial)
 	st->implementation		= IMP_DEFAULT;
 	st->gotArgs				= FALSE;
 	st->gotName				= FALSE;
+	st->nSemicolons			= 0;
 	st->haveQualifyingName	= FALSE;
 	st->argEndPosition		= 0;
 
@@ -970,7 +978,7 @@ static void reinitStatementWithToken (statementInfo *const st,
 	token = activeToken (st);
 	copyToken (token, save);
 	deleteToken (save);
-	++st->tokenIndex;	/* this is quite save becouse current tokenIndex = 0 */
+	++st->tokenIndex;	/* this is quite safe because current tokenIndex = 0 */
 }
 
 static void initStatement (statementInfo *const st, statementInfo *const parent)
@@ -1087,11 +1095,13 @@ static javaKind javaTagKind (const tagType type)
 	javaKind result = JK_UNDEFINED;
 	switch (type)
 	{
-		case TAG_CLASS:     result = JK_CLASS;     break;
-		case TAG_FIELD:     result = JK_FIELD;     break;
-		case TAG_INTERFACE: result = JK_INTERFACE; break;
-		case TAG_METHOD:    result = JK_METHOD;    break;
-		case TAG_PACKAGE:   result = JK_PACKAGE;   break;
+		case TAG_CLASS:      result = JK_CLASS;         break;
+		case TAG_FIELD:      result = JK_FIELD;         break;
+		case TAG_INTERFACE:  result = JK_INTERFACE;     break;
+		case TAG_METHOD:     result = JK_METHOD;        break;
+		case TAG_PACKAGE:    result = JK_PACKAGE;       break;
+		case TAG_ENUM:       result = JK_ENUMERATION;   break;
+		case TAG_ENUMERATOR: result = JK_ENUMERATOR;    break;
 
 		default: Assert ("Bad Java tag type" == NULL); break;
 	}
@@ -1177,6 +1187,7 @@ static const char* accessField (const statementInfo *const st)
 }
 
 static void addOtherFields (tagEntryInfo* const tag, const tagType type,
+							const tokenInfo *const nameToken,
 							const statementInfo *const st, vString *const scope)
 {
 	/*  For selected tag types, append an extension flag designating the
@@ -1187,9 +1198,6 @@ static void addOtherFields (tagEntryInfo* const tag, const tagType type,
 		default: break;
 
 		case TAG_NAMESPACE:
-			/* D nested template block */
-			if (!isLanguage(Lang_d))
-				break;
 		case TAG_CLASS:
 		case TAG_ENUM:
 		case TAG_ENUMERATOR:
@@ -1248,40 +1256,90 @@ static void addOtherFields (tagEntryInfo* const tag, const tagType type,
 		if (((TOKEN_NAME == st->firstToken->type) || isDataTypeKeyword(st->firstToken))
 			&& (0 != strcmp(vStringValue(st->firstToken->name), tag->name)))
 		{
-			tag->extensionFields.varType = getVarType(st);
+			tag->extensionFields.varType = getVarType(st, nameToken);
 		}
 	}
 }
 
-static const char *getVarType (const statementInfo *const st)
+static const char *getVarType (const statementInfo *const st,
+							   const tokenInfo *const nameToken)
 {
 	static vString *vt = NULL;
 	unsigned int i;
+	unsigned int end = st->tokenIndex;
+	boolean seenType = FALSE;
 
-	if (! st->gotArgs)
-		return vStringValue(st->firstToken->name);	/* ignore non-functions */
+	switch (st->declaration) {
+		case DECL_BASE:
+		case DECL_FUNCTION:
+		case DECL_FUNCTION_TEMPLATE:
+			break;
+		default:
+			return vStringValue(st->firstToken->name);
+	}
 
 	if (vt == NULL)
 		vt = vStringNew();
 	else
 		vStringClear(vt);
 
+	/* find the end of the type signature in the token list */
 	for (i = 0; i < st->tokenIndex; i++)
+	{
+		const tokenInfo *const t = st->token[i];
+
+		/* stop if we find the token used to generate the tag name, or
+		 * a name token in the middle yet not preceded by a scope separator */
+		if ((t == nameToken ||
+		     (t->type == nameToken->type &&
+		      t->keyword == nameToken->keyword &&
+		      t->lineNumber == nameToken->lineNumber &&
+		      strcmp(vStringValue(t->name), vStringValue(nameToken->name)) == 0)) ||
+		    (t->type == TOKEN_NAME && seenType &&
+		     (i > 0 && st->token[i - 1]->type != TOKEN_DOUBLE_COLON)))
+		{
+			break;
+		}
+		if (t->type != TOKEN_DOUBLE_COLON)
+			end = i + 1;
+		if (t->type == TOKEN_NAME)
+			seenType = TRUE;
+		else if (t->type == TOKEN_KEYWORD && isDataTypeKeyword(t))
+			seenType = TRUE;
+	}
+
+	/* ugly historic workaround when we can't figure out the type */
+	if (end < 2 && ! st->gotArgs)
+		return vStringValue(st->firstToken->name);
+
+	for (i = 0; i < end; i++)
 	{
 		tokenInfo *t = st->token[i];
 
 		switch (t->type)
 		{
 			case TOKEN_NAME:	/* user typename */
-				if (strcmp(vStringValue(t->name), vStringValue(st->firstToken->name)) != 0)
-					continue;
 				break;
 			case TOKEN_KEYWORD:
-				if (t->keyword != KEYWORD_EXTERN && t->keyword != KEYWORD_STATIC)	/* uninteresting keywords */
+				if ((t->keyword != KEYWORD_EXTERN && t->keyword != KEYWORD_STATIC) &&	/* uninteresting keywords */
+				    (st->gotArgs ||
+				     /* ignore uninteresting keywords for non-functions */
+				     (t->keyword != KEYWORD_PUBLIC &&
+				      t->keyword != KEYWORD_PRIVATE &&
+				      t->keyword != KEYWORD_PROTECTED &&
+				      t->keyword != KEYWORD_FINAL &&
+				      t->keyword != KEYWORD_TYPEDEF &&
+				      /* hack for D static conditions */
+				      t->keyword != KEYWORD_IF)))
+				{
 					break;
+				}
 				continue;
 			case TOKEN_STAR: vStringCatS(vt, " *"); continue;
 			case TOKEN_ARRAY: vStringCatS(vt, "[]"); continue;
+			case TOKEN_DOUBLE_COLON:
+				vStringCatS(vt, "::");
+				continue;
 			default: continue;
 		}
 		if (vStringLength(vt) > 0)
@@ -1420,7 +1478,7 @@ static void makeTag (const tokenInfo *const token,
 		e.type = type;
 
 		findScopeHierarchy (scope, st);
-		addOtherFields (&e, type, st, scope);
+		addOtherFields (&e, type, token, st, scope);
 
 #ifdef DEBUG_C
 		printTagEntry(&e);
@@ -1642,6 +1700,16 @@ static void skipToMatch (const char *const pair)
 				break;
 			}
 		}
+		/* early out if matching "<>" and we encounter a ";" or "{" to mitigate
+		 * match problems with C++ generics containing a static expression like
+		 *     foo<X<Y> bar;
+		 * normally neither ";" nor "{" could appear inside "<>" anyway. */
+		else if (isLanguage (Lang_cpp) && begin == '<' &&
+		         (c == ';' || c == '{'))
+		{
+			cppUngetc (c);
+			break;
+		}
 	}
 	if (c == EOF)
 	{
@@ -1678,9 +1746,12 @@ static keywordId analyzeKeyword (const char *const name)
 {
 	const keywordId id = (keywordId) lookupKeyword (name, getSourceLanguage ());
 
-	/* ignore D @attributes, but show them in function signatures */
-	if (isLanguage(Lang_d) && id == KEYWORD_NONE && name[0] == '@')
+	/* ignore D @attributes and Java @annotations(...), but show them in function signatures */
+	if ((isLanguage(Lang_d) || isLanguage(Lang_java)) && id == KEYWORD_NONE && name[0] == '@')
+	{
+		skipParens(); /* if annotation has parameters, skip them */
 		return KEYWORD_CONST;
+	}
 	return id;
 }
 
@@ -2011,6 +2082,7 @@ static void processToken (tokenInfo *const token, statementInfo *const st)
 		case KEYWORD_SHORT:		st->declaration = DECL_BASE;		break;
 		case KEYWORD_SIGNED:	st->declaration = DECL_BASE;		break;
 		case KEYWORD_STRUCT:	checkIsClassEnum (st, DECL_STRUCT);	break;
+		case KEYWORD_STATIC_ASSERT: skipParens ();                  break;
 		case KEYWORD_THROWS:	discardTypeList (token);			break;
 		case KEYWORD_TYPEDEF:	st->scope	= SCOPE_TYPEDEF;		break;
 		case KEYWORD_UNION:		st->declaration = DECL_UNION;		break;
@@ -2220,6 +2292,7 @@ static boolean skipPostArgumentStuff (statementInfo *const st,
 					case KEYWORD_ATTRIBUTE:	skipParens ();	break;
 					case KEYWORD_THROW:	skipParens ();		break;
 					case KEYWORD_CONST:						break;
+					case KEYWORD_NOEXCEPT:					break;
 					case KEYWORD_TRY:						break;
 					case KEYWORD_VOLATILE:					break;
 
@@ -2240,7 +2313,13 @@ static boolean skipPostArgumentStuff (statementInfo *const st,
 						break;
 
 					default:
-						if (isType (token, TOKEN_NONE))
+						/* "override" and "final" are only keywords in the declaration of a virtual
+						 * member function, so need to be handled specially, not as keywords */
+						if (isLanguage(Lang_cpp) && isType (token, TOKEN_NAME) &&
+							(strcmp ("override", vStringValue (token->name)) == 0 ||
+							 strcmp ("final", vStringValue (token->name)) == 0))
+							;
+						else if (isType (token, TOKEN_NONE))
 							;
 						else if (info->isKnrParamList  &&  info->parameterCount > 0)
 							++elementCount;
@@ -2598,14 +2677,21 @@ static void addContext (statementInfo *const st, const tokenInfo* const token)
 
 static boolean inheritingDeclaration (declType decl)
 {
-	return (boolean) (decl == DECL_CLASS ||
-					  decl == DECL_STRUCT ||
-					  decl == DECL_INTERFACE);
+	/* enum base types */
+	if (decl == DECL_ENUM)
+	{
+		return (boolean) (isLanguage (Lang_cpp) || isLanguage (Lang_csharp) ||
+			isLanguage (Lang_d));
+	}
+	return (boolean) (
+		decl == DECL_CLASS ||
+		decl == DECL_STRUCT ||
+		decl == DECL_INTERFACE);
 }
 
 static void processColon (statementInfo *const st)
 {
-	int c = skipToNonWhite ();
+	int c = cppGetc ();
 	const boolean doubleColon = (boolean) (c == ':');
 
 	if (doubleColon)
@@ -2616,9 +2702,8 @@ static void processColon (statementInfo *const st)
 	else
 	{
 		cppUngetc (c);
-		if (((isLanguage (Lang_cpp) &&
-				(st->declaration == DECL_CLASS || st->declaration == DECL_STRUCT)) ||
-			isLanguage (Lang_csharp) || isLanguage (Lang_d) || isLanguage (Lang_vala)) &&
+		if ((isLanguage (Lang_cpp) || isLanguage (Lang_csharp) || isLanguage (Lang_d) ||
+			isLanguage (Lang_vala)) &&
 			inheritingDeclaration (st->declaration))
 		{
 			readParents (st, ':');
@@ -2631,15 +2716,6 @@ static void processColon (statementInfo *const st)
 			else if (c == ';')
 				setToken (st, TOKEN_SEMICOLON);
 		}
-		else if (isLanguage (Lang_cpp) && st->declaration == DECL_ENUM)
-		{
-			/* skip enum's base type */
-			c = skipToOneOf ("{;");
-			if (c == '{')
-				setToken (st, TOKEN_BRACE_OPEN);
-			else if (c == ';')
-				setToken (st, TOKEN_SEMICOLON);
-		}
 		else
 		{
 			const tokenInfo *const prev  = prevToken (st, 1);
@@ -2779,6 +2855,9 @@ static void nextToken (statementInfo *const st)
 		}
 	} while (isType (token, TOKEN_NONE));
 
+	if (isType (token, TOKEN_SEMICOLON) && st->parent)
+		st->parent->nSemicolons ++;
+
 	/* We want to know about non-keyword variable types */
 	if (TOKEN_NONE == st->firstToken->type)
 	{
@@ -2906,7 +2985,9 @@ static void tagCheck (statementInfo *const st)
 	{
 		case TOKEN_NAME:
 		{
-			if (insideEnumBody (st))
+			if (insideEnumBody (st) &&
+				/* Java enumerations can contain members after a semicolon */
+				(! isLanguage(Lang_java) || st->parent->nSemicolons < 1))
 				qualifyEnumeratorTag (st, token);
 			break;
 		}
@@ -2945,7 +3026,15 @@ static void tagCheck (statementInfo *const st)
 				tokenInfo *name_token = (tokenInfo *)prev;
 				boolean free_name_token = FALSE;
 
-				if (isType (name_token, TOKEN_NAME))
+				/* C++ 11 allows class <name> final { ... } */
+				if (isLanguage (Lang_cpp) && isType (prev, TOKEN_NAME) &&
+					strcmp("final", vStringValue(prev->name)) == 0 &&
+					isType(prev2, TOKEN_NAME))
+				{
+					name_token = (tokenInfo *)prev2;
+					copyToken (st->blockName, name_token);
+				}
+				else if (isType (name_token, TOKEN_NAME))
 				{
 					if (!isLanguage (Lang_vala))
 						copyToken (st->blockName, name_token);
@@ -2970,13 +3059,6 @@ static void tagCheck (statementInfo *const st)
 						}
 					}
 				}
-				/* C++ 11 allows class <name> final { ... } */
-				else if (isLanguage (Lang_cpp) && isType (prev, TOKEN_KEYWORD) &&
-						 prev->keyword == KEYWORD_FINAL && isType(prev2, TOKEN_NAME))
-				{
-					name_token = (tokenInfo *)prev2;
-					copyToken (st->blockName, name_token);
-				}
 				else if (isLanguage (Lang_csharp))
 					makeTag (prev, st, FALSE, TAG_PROPERTY);
 				else
@@ -3015,7 +3097,9 @@ static void tagCheck (statementInfo *const st)
 		case TOKEN_SEMICOLON:
 		case TOKEN_COMMA:
 		{
-			if (insideEnumBody (st))
+			if (insideEnumBody (st) &&
+				/* Java enumerations can contain members after a semicolon */
+				(! isLanguage (Lang_java) || st->parent->nSemicolons < 2))
 				;
 			else if (isType (prev, TOKEN_NAME))
 			{
@@ -3088,7 +3172,7 @@ static boolean findCTags (const unsigned int passCount)
 	contextual_fake_count = 0;
 
 	Assert (passCount < 3);
-	cppInit ((boolean) (passCount > 1), isLanguage (Lang_csharp));
+	cppInit ((boolean) (passCount > 1), isLanguage (Lang_csharp), isLanguage (Lang_cpp));
 
 	exception = (exception_t) setjmp (Exception);
 	retry = FALSE;
