@@ -33,7 +33,17 @@ static kindOption PythonKinds[] = {
 	{TRUE, 'f', "function", "functions"},
 	{TRUE, 'm', "method",   "class methods"},
     {TRUE, 'v', "variable", "variables"},
-    {TRUE, 'i', "namespace", "imports"}
+    /* defined as externvar to get those excluded as forward type in symbols.c:goto_tag()
+     * so we can jump to the real implementation (if known) instead of to the import statement */
+    {TRUE, 'x', "externvar", "imports"}
+};
+
+typedef enum {
+	A_PUBLIC, A_PRIVATE, A_PROTECTED
+} pythonAccess;
+
+static const char *const PythonAccesses[] = {
+	"public", "private", "protected"
 };
 
 static char const * const singletriple = "'''";
@@ -43,8 +53,6 @@ static char const * const doubletriple = "\"\"\"";
 *   FUNCTION DEFINITIONS
 */
 
-#define vStringLast(vs) ((vs)->buffer[(vs)->length - 1])
-
 static boolean isIdentifierFirstCharacter (int c)
 {
 	return (boolean) (isalpha (c) || c == '_');
@@ -78,6 +86,47 @@ static const char *get_class_name_from_parent (const char *parent)
 	return result;
 }
 
+/* follows PEP-8, and always reports single-underscores as protected
+ * See:
+ * - http://www.python.org/dev/peps/pep-0008/#method-names-and-instance-variables
+ * - http://www.python.org/dev/peps/pep-0008/#designing-for-inheritance
+ */
+static pythonAccess accessFromIdentifier (const vString *const ident,
+	pythonKind kind, boolean has_parent, boolean parent_is_class)
+{
+	const char *const p = vStringValue (ident);
+	const size_t len = vStringLength (ident);
+
+	/* inside a function/method, private */
+	if (has_parent && !parent_is_class)
+		return A_PRIVATE;
+	/* not starting with "_", public */
+	else if (len < 1 || p[0] != '_')
+		return A_PUBLIC;
+	/* "__...__": magic methods */
+	else if (kind == K_METHOD && parent_is_class &&
+			 len > 3 && p[1] == '_' && p[len - 2] == '_' && p[len - 1] == '_')
+		return A_PUBLIC;
+	/* "__...": name mangling */
+	else if (parent_is_class && len > 1 && p[1] == '_')
+		return A_PRIVATE;
+	/* "_...": suggested as non-public, but easily accessible */
+	else
+		return A_PROTECTED;
+}
+
+static void addAccessFields (tagEntryInfo *const entry,
+	const vString *const ident, pythonKind kind,
+	boolean has_parent, boolean parent_is_class)
+{
+	pythonAccess access;
+
+	access = accessFromIdentifier (ident, kind, has_parent, parent_is_class);
+	entry->extensionFields.access = PythonAccesses [access];
+	/* FIXME: should we really set isFileScope in addition to access? */
+	if (access == A_PRIVATE)
+		entry->isFileScope = TRUE;
+}
 
 /* Given a string with the contents of a line directly after the "def" keyword,
  * extract all relevant information and create a tag.
@@ -105,30 +154,19 @@ static void makeFunctionTag (vString *const function,
 		{
 			tag.kindName = PythonKinds[K_METHOD].name;
 			tag.kind = PythonKinds[K_METHOD].letter;
-			tag.extensionFields.scope [0] = "class";
+			tag.extensionFields.scope [0] = PythonKinds[K_CLASS].name;
 			tag.extensionFields.scope [1] = vStringValue (parent);
 		}
 		else
 		{
-			tag.extensionFields.scope [0] = "function";
+			tag.extensionFields.scope [0] = PythonKinds[K_FUNCTION].name;
 			tag.extensionFields.scope [1] = vStringValue (parent);
 		}
 	}
 
-	/* If a function starts with __, we mark it as file scope.
-	 * FIXME: What is the proper way to signal such attributes?
-	 * TODO: What does functions/classes starting with _ and __ mean in python?
-	 */
-	if (strncmp (vStringValue (function), "__", 2) == 0 &&
-		strcmp (vStringValue (function), "__init__") != 0)
-	{
-		tag.extensionFields.access = "private";
-		tag.isFileScope = TRUE;
-	}
-	else
-	{
-		tag.extensionFields.access = "public";
-	}
+	addAccessFields (&tag, function, is_class_parent ? K_METHOD : K_FUNCTION,
+		vStringLength (parent) > 0, is_class_parent);
+
 	makeTagEntry (&tag);
 }
 
@@ -146,20 +184,23 @@ static void makeClassTag (vString *const class, vString *const inheritance,
 	{
 		if (is_class_parent)
 		{
-			tag.extensionFields.scope [0] = "class";
+			tag.extensionFields.scope [0] = PythonKinds[K_CLASS].name;
 			tag.extensionFields.scope [1] = vStringValue (parent);
 		}
 		else
 		{
-			tag.extensionFields.scope [0] = "function";
+			tag.extensionFields.scope [0] = PythonKinds[K_FUNCTION].name;
 			tag.extensionFields.scope [1] = vStringValue (parent);
 		}
 	}
 	tag.extensionFields.inheritance = vStringValue (inheritance);
+	addAccessFields (&tag, class, K_CLASS, vStringLength (parent) > 0,
+		is_class_parent);
 	makeTagEntry (&tag);
 }
 
-static void makeVariableTag (vString *const var, vString *const parent)
+static void makeVariableTag (vString *const var, vString *const parent,
+	boolean is_class_parent)
 {
 	tagEntryInfo tag;
 	initTagEntry (&tag, vStringValue (var));
@@ -167,9 +208,11 @@ static void makeVariableTag (vString *const var, vString *const parent)
 	tag.kind = PythonKinds[K_VARIABLE].letter;
 	if (vStringLength (parent) > 0)
 	{
-		tag.extensionFields.scope [0] = "class";
+		tag.extensionFields.scope [0] = PythonKinds[K_CLASS].name;
 		tag.extensionFields.scope [1] = vStringValue (parent);
 	}
+	addAccessFields (&tag, var, K_VARIABLE, vStringLength (parent) > 0,
+		is_class_parent);
 	makeTagEntry (&tag);
 }
 
@@ -196,25 +239,35 @@ static const char *skipEverything (const char *cp)
 	int match;
 	for (; *cp; cp++)
 	{
+		if (*cp == '#')
+			return strchr(cp, '\0');
+
 		match = 0;
-		if (*cp == '"' || *cp == '\'' || *cp == '#')
+		if (*cp == '"' || *cp == '\'')
 			match = 1;
 
 		/* these checks find unicode, binary (Python 3) and raw strings */
-		if (!match && (
-			!strncasecmp(cp, "u'", 2) || !strncasecmp(cp, "u\"", 2) ||
-			!strncasecmp(cp, "r'", 2) || !strncasecmp(cp, "r\"", 2) ||
-			!strncasecmp(cp, "b'", 2) || !strncasecmp(cp, "b\"", 2)))
+		if (!match)
 		{
-			match = 1;
-			cp += 1;
-		}
-		if (!match && (
-			!strncasecmp(cp, "ur'", 3) || !strncasecmp(cp, "ur\"", 3) ||
-			!strncasecmp(cp, "br'", 3) || !strncasecmp(cp, "br\"", 3)))
-		{
-			match = 1;
-			cp += 2;
+			boolean r_first = (*cp == 'r' || *cp == 'R');
+
+			/* "r" | "R" | "u" | "U" | "b" | "B" */
+			if (r_first || *cp == 'u' || *cp == 'U' ||  *cp == 'b' || *cp == 'B')
+			{
+				unsigned int i = 1;
+
+				/*  r_first -> "rb" | "rB" | "Rb" | "RB"
+				   !r_first -> "ur" | "UR" | "Ur" | "uR" | "br" | "Br" | "bR" | "BR" */
+				if (( r_first && (cp[i] == 'b' || cp[i] == 'B')) ||
+					(!r_first && (cp[i] == 'r' || cp[i] == 'R')))
+					i++;
+
+				if (cp[i] == '\'' || cp[i] == '"')
+				{
+					match = 1;
+					cp += i;
+				}
+			}
 		}
 		if (match)
 		{
@@ -223,6 +276,8 @@ static const char *skipEverything (const char *cp)
 		}
 		if (isIdentifierFirstCharacter ((int) *cp))
 			return cp;
+		if (match)
+			cp--; /* avoid jumping over the character after a skipped string */
 	}
 	return cp;
 }
@@ -418,10 +473,8 @@ static boolean constructParentString(NestingLevels *nls, int indent,
 	return is_class;
 }
 
-/* Check whether parent's indentation level is higher than the current level and
- * if so, remove it.
- */
-static void checkParent(NestingLevels *nls, int indent, vString *parent)
+/* Check indentation level and truncate nesting levels accordingly */
+static void checkIndent(NestingLevels *nls, int indent)
 {
 	int i;
 	NestingLevel *n;
@@ -429,14 +482,10 @@ static void checkParent(NestingLevels *nls, int indent, vString *parent)
 	for (i = 0; i < nls->n; i++)
 	{
 		n = nls->levels + i;
-		/* is there a better way to compare two vStrings? */
-		if (n && strcmp(vStringValue(parent), vStringValue(n->name)) == 0)
+		if (n && indent <= n->indentation)
 		{
-			if (indent <= n->indentation)
-			{
-				/* remove this level by clearing its name */
-				vStringClear(n->name);
-			}
+			/* truncate levels */
+			nls->n = i;
 			break;
 		}
 	}
@@ -493,6 +542,7 @@ static char const *find_triple_start(char const *string, char const **which)
 			}
 			cp = skipString(cp);
 			if (!*cp) break;
+			cp--; /* avoid jumping over the character after a skipped string */
 		}
 	}
 	return NULL;
@@ -576,7 +626,15 @@ static const char *skipTypeDecl (const char *cp, boolean *is_class)
 	}
 	/* limit so that we don't pick off "int item=obj()" */
 	while (*ptr && loopCount++ < 2) {
-		while (*ptr && *ptr != '=' && *ptr != '(' && !isspace(*ptr)) ptr++;
+		while (*ptr && *ptr != '=' && *ptr != '(' && !isspace(*ptr)) {
+			/* skip over e.g. 'cpdef numpy.ndarray[dtype=double, ndim=1]' */
+			if(*ptr == '[') {
+				while (*ptr && *ptr != ']') ptr++;
+				if (*ptr) ptr++;
+			} else {
+				ptr++;
+			}
+		}
 		if (!*ptr || *ptr == '=') return NULL;
 		if (*ptr == '(') {
 		    return lastStart; /* if we stopped on a '(' we are done */
@@ -695,7 +753,39 @@ static void findPythonTags (void)
 			continue;
 		}
 		
-		checkParent(nesting_levels, indent, parent);
+		checkIndent(nesting_levels, indent);
+
+		/* Find global and class variables */
+		variable = findVariable(line);
+		if (variable)
+		{
+			const char *start = variable;
+			char *arglist;
+			boolean parent_is_class;
+
+			vStringClear (name);
+			while (isIdentifierCharacter ((int) *start))
+			{
+				vStringPut (name, (int) *start);
+				++start;
+			}
+			vStringTerminate (name);
+
+			parent_is_class = constructParentString(nesting_levels, indent, parent);
+			if (varIsLambda (variable, &arglist))
+			{
+				/* show class members or top-level script lambdas only */
+				if (parent_is_class || vStringLength(parent) == 0)
+					makeFunctionTag (name, parent, parent_is_class, arglist);
+				eFree (arglist);
+			}
+			else
+			{
+				/* skip variables in methods */
+				if (parent_is_class || vStringLength(parent) == 0)
+					makeVariableTag (name, parent, parent_is_class);
+			}
+		}
 
 		/* Deal with multiline string start. */
 		longstring = find_triple_start(cp, &longStringLiteral);
@@ -757,40 +847,6 @@ static void findPythonTags (void)
 				addNestingLevel(nesting_levels, indent, name, is_class);
 			}
 		}
-		/* Find global and class variables */
-		variable = findVariable(line);
-		if (variable)
-		{
-			const char *start = variable;
-			char *arglist;
-			boolean parent_is_class;
-
-			vStringClear (name);
-			while (isIdentifierCharacter ((int) *start))
-			{
-				vStringPut (name, (int) *start);
-				++start;
-			}
-			vStringTerminate (name);
-
-			parent_is_class = constructParentString(nesting_levels, indent, parent);
-			if (varIsLambda (variable, &arglist))
-			{
-				/* show class members or top-level script lambdas only */
-				if (parent_is_class || vStringLength(parent) == 0)
-					makeFunctionTag (name, parent, parent_is_class, arglist);
-				if (arglist != NULL)
-					eFree (arglist);
-			}
-			else
-			{
-				/* skip variables in methods */
-				if (! parent_is_class && vStringLength(parent) > 0)
-					continue;
-
-				makeVariableTag (name, parent);
-			}
-		}
 		/* Find and parse imports */
 		parseImports(line);
 	}
